/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "operator/operator.pmh"

using namespace fe;

void SurfaceBindOp::initialize(void)
{
	catalog<bool>("Bind Primitives")=1;
	catalog<String>("Bind Primitives","hint")=
			"Affect whole primitives instead of individual points";

	catalog<I32>("Bindings")=1;
	catalog<I32>("Bindings","min")=1;
	catalog<String>("Bindings","hint")=
			"Maximum number of attachment contacts";

	catalog<String>("Bind Attribute Prefix")="bind";
	catalog<String>("Bind Attribute Prefix","hint")=
			"Prefix of attribute names added to output surface describing"
			" where each point on the input surface is attached to"
			" the associated driver";

	catalog< sp<Component> >("Input Surface");
	catalog< sp<Component> >("Driver Surface");
}

void SurfaceBindOp::handle(Record& a_rSignal)
{
//	feLog("SurfaceBindOp::handle\n");

	//* access output surface (copy of input surface)
	sp<SurfaceAccessibleI> spOutputAccessible=surfaceOutput(a_rSignal);
	if(spOutputAccessible.isNull())
	{
		feLog("SurfaceBindOp::handle"
				" output surface not SurfaceAccessibleI\n");
		catalog<String>("error")="inaccessible output";
		return;
	}

	//* access reference curves
	sp<SurfaceAccessibleI> spDriverAccessible=
			catalog< sp<Component> >("Driver Surface");
	if(spDriverAccessible.isNull())
	{
		feLog("SurfaceBindOp::handle"
				" driver surface not SurfaceAccessibleI\n");
		catalog<String>("error")="inaccessible driver surface";
		return;
	}
	sp<SurfaceI> spDriver=spDriverAccessible->surface();
	if(spDriver.isNull())
	{
		feLog("SurfaceBindOp::handle"
				" driver surface failed to provide SurfaceI\n");
		catalog<String>("error")="invalid driver surface";
		return;
	}
	spDriver->setRefinement(0);

	const bool bindPrimitives=catalog<bool>("Bind Primitives");
	const SurfaceAccessibleI::Element element=bindPrimitives?
			SurfaceAccessibleI::e_primitive:
			SurfaceAccessibleI::e_point;
	const SurfaceAccessibleI::Attribute attribute=bindPrimitives?
			SurfaceAccessibleI::e_vertices:
			SurfaceAccessibleI::e_position;

	sp<SurfaceAccessorI> spOutputElements=
			spOutputAccessible->accessor(element,attribute);
	FEASSERT(spOutputElements.isValid());

	std::vector< sp<SurfaceAccessorI> > outputFaces;
	std::vector< sp<SurfaceAccessorI> > outputBarys;
	const I32 bindings=accessBindings(spOutputAccessible,
			outputFaces,outputBarys,TRUE);
	if(bindings<1)
	{
		return;
	}

	I32 face= -1;
	Barycenter<Real> barycenter;

	const U32 count=spOutputElements->count();
	for(U32 index=0;index<count;index++)
	{
		const U32 subCount=spOutputElements->subCount(index);
		if(subCount<1)
		{
			continue;
		}

		const SpatialVector point=spOutputElements->spatialVector(index,0);
		std::vector< sp<SurfaceI::ImpactI> > impactArray=
				spDriver->nearestPoints(point,bindings);
		const U32 impacts=impactArray.size();

		for(U32 binding=0;binding<bindings;binding++)
		{
			if(binding<impacts && impactArray[binding].isValid())
			{
				sp<SurfaceI::ImpactI>& rspImpact=impactArray[binding];

				face=rspImpact->face();
				barycenter=rspImpact->barycenter();
			}
			else
			{
				face= -1;
				set(barycenter);
			}

			outputFaces[binding]->set(index,face);
			outputBarys[binding]->set(index,barycenter);
		}
	}

//	feLog("SurfaceBindOp::handle done\n");
}

I32 SurfaceBindOp::accessBindings(sp<SurfaceAccessibleI>& a_rspAccessibleI,
	std::vector< sp<SurfaceAccessorI> >&a_rFaces,
	std::vector< sp<SurfaceAccessorI> >&a_rBarys,
	BWORD a_allowCreation)
{
	const I32 bindings=catalog<I32>("Bindings");
	const String bindPrefix=catalog<String>("Bind Attribute Prefix");

	const bool bindPrimitives=catalog<bool>("Bind Primitives");
	const SurfaceAccessibleI::Element element=bindPrimitives?
			SurfaceAccessibleI::e_primitive:
			SurfaceAccessibleI::e_point;

	a_rFaces.resize(bindings);
	a_rBarys.resize(bindings);
	U32 count=0;
	for(U32 binding=0;binding<bindings;binding++)
	{
		String faceName;
		String baryName;
		faceName.sPrintf("%sFace%d",bindPrefix.c_str(),binding);
		baryName.sPrintf("%sBary%d",bindPrefix.c_str(),binding);

		a_rFaces[binding]=a_rspAccessibleI->accessor(element,
				faceName,a_allowCreation);
		if(a_rFaces[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp::handle face attribute"
						" \"%s\" not accessible\n",faceName.c_str());
				catalog<String>("error")="inaccessible face attribute \""+
							faceName+"\"";
				return -1;
			}
			break;
		}

		a_rBarys[binding]=a_rspAccessibleI->accessor(element,
				baryName,a_allowCreation);
		if(a_allowCreation && a_rBarys[binding].isNull())
		{
			if(a_allowCreation)
			{
				feLog("SurfaceBindOp::handle barycentric attribute"
						" \"%s\" not accessible\n",baryName.c_str());
				catalog<String>("error")=
							"inaccessible barycentric attribute \""+
							baryName+"\"";
				return -1;
			}
			break;
		}

		count++;
	}

	return count;
}
