/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define FE_SWO_DEBUG	FALSE

#include "operator/operator.pmh"

using namespace fe;

void SurfaceWrapOp::initialize(void)
{
	//* NOTE inherits attributes from SurfaceBindOp

	catalog<String>("Bind Attribute Prefix","hint")=
			"Prefix of attribute names on input surface describing"
			" where each point on the input surface is attached to"
			" the associated driver";

	catalog<Real>("Distance Bias")=0.0;
	catalog<Real>("Distance Bias","min")=0.0;
	catalog<Real>("Distance Bias","max")=10.0;
	catalog<String>("Distance Bias","hint")=
			"Significance of more distant faces towards weighting";

	catalog< sp<Component> >("Deformed Surface");

	catalog< sp<Component> >("Output Surface");
	catalog<String>("Output Surface","IO")="output";
	catalog<String>("Output Surface","copy")="Input Surface";
//	catalog<bool>("Output Surface","recycle")=true;
}

void SurfaceWrapOp::handle(Record& a_rSignal)
{
#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle\n");
#endif

	const String bindPrefix=catalog<String>("Bind Attribute Prefix");
	m_faceName=bindPrefix+"Face0";
	m_baryName=bindPrefix+"Bary0";

	//* access output curves (copy of input surface)
	m_spOutputAccessible=surfaceOutput(a_rSignal);
	if(m_spOutputAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" output surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access output data";
		return;
	}

	//* access reference surface
	m_spInputAccessible=catalog< sp<Component> >("Input Surface");
	if(m_spInputAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" reference surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access reference surface";
		return;
	}

	//* access original driver
	m_spDriverAccessible=catalog< sp<Component> >("Driver Surface");
	if(m_spDriverAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" driver surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access driver surface";
		return;
	}

	//* access deformed driver
	m_spDeformedAccessible=catalog< sp<Component> >("Deformed Surface");
	if(m_spDeformedAccessible.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" deformed surface not SurfaceAccessibleI\n");
		catalog<String>("error")="can't access deformed curves";
		return;
	}

	m_spDriver=m_spDriverAccessible->surface();
	if(m_spDriver.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" driver surface failed to provide SurfaceI\n");
		catalog<String>("error")="invalid driver surface";
		return;
	}
	m_spDriver->setRefinement(0);
	m_spDriver->center();

	m_spDeformed=m_spDeformedAccessible->surface();
	if(m_spDeformed.isNull())
	{
		feLog("SurfaceWrapOp::handle"
				" deformed surface failed to provide SurfaceI\n");
		catalog<String>("error")="invalid deformed surface";
		return;
	}
	m_spDeformed->setRefinement(0);
	m_spDeformed->center();

	const bool bindPrimitives=catalog<bool>("Bind Primitives");
	m_element=bindPrimitives?
			SurfaceAccessibleI::e_primitive:
			SurfaceAccessibleI::e_point;
	m_attribute=bindPrimitives?
			SurfaceAccessibleI::e_vertices:
			SurfaceAccessibleI::e_position;

	sp<SurfaceAccessorI> spOutputElements=
			m_spOutputAccessible->accessor(m_element,m_attribute);
	FEASSERT(spOutputElements.isValid());

	const U32 primitiveCount=spOutputElements->count();

	setFullRange(0,primitiveCount-1);
	OperatorThreaded::handle(a_rSignal);

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::handle done\n");
#endif
}

void SurfaceWrapOp::run(I32 a_id,Range a_range)
{
	const I32 start=a_range.m_start;
	const I32 end=a_range.m_end;

	const Real distanceBias=catalog<Real>("Distance Bias");

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run running thread %d [%d,%d]\n",a_id,start,end);
#endif

	sp<SurfaceAccessorI> spInputElements=
			m_spInputAccessible->accessor(m_element,m_attribute);
	FEASSERT(spInputElements.isValid());

	sp<SurfaceAccessorI> spOutputElements=
			m_spOutputAccessible->accessor(m_element,m_attribute);
	FEASSERT(spOutputElements.isValid());

	std::vector< sp<SurfaceAccessorI> > outputFaces;
	std::vector< sp<SurfaceAccessorI> > outputBarys;
	const I32 bindings=accessBindings(m_spOutputAccessible,
			outputFaces,outputBarys,FALSE);
	if(bindings<1)
	{
		return;
	}

	for(U32 index=start;index<=end;index++)
	{
		const U32 subCount=spInputElements->subCount(index);
		if(subCount<1)
		{
			continue;
		}

		SpatialVector point[subCount];
		SpatialVector result[subCount];
		Real totalWeight[subCount];
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			point[subIndex]=spInputElements->spatialVector(index,subIndex);
			set(result[subIndex]);
			totalWeight[subIndex]=0.0;
		}

		for(U32 binding=0;binding<bindings;binding++)
		{
			const I32 face=outputFaces[binding]->integer(index);
			if(face<0)
			{
				continue;
			}

			const Barycenter<Real> barycenter=
					outputBarys[binding]->spatialVector(index);
			const SpatialTransform xformDriver=
					m_spDriver->sample(face,barycenter);
			const SpatialTransform xformDeformed=
					m_spDeformed->sample(face,barycenter);
			SpatialTransform invDriver;
			invert(invDriver,xformDriver);
			const SpatialTransform conversion=invDriver*xformDeformed;

			const Real distance=
					magnitude(xformDriver.translation()-point[0]);
			const Real weight=1.0/(distance+distanceBias+1e-9);

			for(U32 subIndex=0;subIndex<subCount;subIndex++)
			{
				SpatialVector transformed;
				transformVector(conversion,point[subIndex],transformed);
				result[subIndex]+=weight*transformed;
				totalWeight[subIndex]+=weight;
			}
		}

		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			if(totalWeight[subIndex]>0.0)
			{
				spOutputElements->set(index,subIndex,
						result[subIndex]/totalWeight[subIndex]);
			}
		}
	}

#if FE_SWO_DEBUG
	feLog("SurfaceWrapOp::run done running thread %d\n",a_id);
#endif
}
